#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : radarsat.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import re
import copy
import logging
import numpy as np
from tqdm import tqdm

from torchsar.products.ceos import decfmtfceos
from torchsar.products.utils import splitfmt
from torchsar.products.record import readrcd, readrcd1item
from torchsar.sarcfg.sensor_satellite import SENSOR_SATELLITE

# B: number in binary format
# A: string
# I: char
SarDataFileFileDescriptorRecordRADARSAT = {
    # SAR DATA FILE - FILE DESCRIPTOR RECORD (FIXED SEGMENT)
    'Record sequence number': [(1, 4), '1B4', 0],
    
}

SarDataFileSignalDataRecordRADARSAT = {
    'Record sequence number': [(1, 4), '1B4', 0],
}


def _getdtype_component(fmtrcd, fmtuser):

    raise(ImportError('Not opened yet!'))



def read_radarsat_sar_raw(filepath, sl=1, el=-1, rmbp=False):
    r"""read RADARSAT SAR raw data

    read RADARSAT raw data pulse from line :attr:`sl` to line :attr:`el`. This function call
    function :func:`read_ceos_sar_raw` firstly and do some post-process.


    Parameters
    ----------
    filepath : str
        RADARSAT SAR raw data file path string with format ``.001``, such as ``DAT_01.001``.
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array
        SAR raw data.

    """

    sensor_name = 'RADARSAT1'
    logging.info("===In read_radarsat_sar_raw...")

    raise(ImportError('Not opened yet!'))



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : alos.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#


import re
import copy
import logging
import numpy as np
from tqdm import tqdm
import torchsar as ts
import torchbox as tb
from torchsar.products.ceos import decfmtfceos
from torchsar.products.utils import splitfmt
from torchsar.products.record import readrcd, readrcd1item, printrcd
from torchsar.sarcfg.sensor_satellite import SENSOR_SATELLITE


LeaderFileImportantImagingParametersRecordALOS = {
    'Scene identifier': [(720 + 21, 720 + 52), '1A32', 0],
    'Ellipsoid designator = "GRS80bbbbbbbbbbb":Fixed': [(720 + 165, 720 + 180), '1A16', 0],
    'Mass of earth (10^24kg)': [(720 + 213, 720 + 228), '1F16', 0],

}

# B: number in binary format
# A: string
# I: char
SarDataFileFileDescriptorRecordALOSPALSAR = {
    # SAR DATA FILE - FILE DESCRIPTOR RECORD (FIXED SEGMENT)
    'Record sequence number': [(1, 4), '1B4', 0],

}

SarDataFileSignalDataRecordALOSPALSAR = {  # level1.0
    'Record sequence number': [(1, 4), '1B4', 0],

}


# B: number in binary format
# A: string
# I: char
SarImageFileFileDescriptorRecordALOSPALSAR = {
    # SAR DATA FILE - FILE DESCRIPTOR RECORD (FIXED SEGMENT)
    'Record sequence number': [(1, 4), '1B4', 0],
}

SarImageFileSignalDataRecordALOSPALSAR = {  # level1.1
    'Record sequence number': [(1, 4), '1B4', 0],
}

SarImageFileProcessedDataRecordALOSPALSAR = {  # level1.5
    'Record sequence number': [(1, 4), '1B4', 0],
}


def _getdtype_component(fmtrcd, fmtuser):
    raise(ImportError('Not opened yet!'))


def get_alos_palsar_plat_position(D):
    """get platform position data

    return a numpy array
        1st: X Y Z Vx Vy Vz
        2st: X Y Z Vx Vy Vz
           :
        nst: X Y Z Vx Vy Vz

    Parameters
    ----------
    D : dict
        LeaderFileImportantImagingParametersRecordALOS dict
    """

    logging.info("===In get_alos_palsar_plat_position...")

    P = []

    P.append(D['1st data point position vector as (X, Y, Z) co-ordinates for spaceborne sensor platform in a reference system (X) (meters)'][2][0])
    P.append(D['1st data point position vector as (X, Y, Z) co-ordinates for spaceborne sensor platform in a reference system (Y) (meters)'][2][0])
    P.append(D['1st data point position vector as (X, Y, Z) co-ordinates for spaceborne sensor platform in a reference system (Z) (meters)'][2][0])
    P.append(D['1st data point velocity vector (Vx, Vy, Vz) in a reference system (Vx) [m/sec]'][2][0])
    P.append(D['1st data point velocity vector (Vx, Vy, Vz) in a reference system (Vy) [m/sec]'][2][0])
    P.append(D['1st data point velocity vector (Vx, Vy, Vz) in a reference system (Vz) [m/sec]'][2][0])
    P = P + D['2nd, 3rd, ... data point position & velocity vectors'][2]

    nPlatformPositionPoints = D['Number of data points'][2][0]
    P = np.array(P)
    P = np.reshape(P, (nPlatformPositionPoints, 6))

    logging.info("===Out get_alos_palsar_plat_position.")
    return P


def get_alos_palsar_attitude(D):
    """get attitude data

    return a numpy array
        1st: Pitch Roll Yaw PitchRate RollRate YawRate
        2st: Pitch Roll Yaw PitchRate RollRate YawRate
           :
        nst: Pitch Roll Yaw PitchRate RollRate YawRate

    Parameters
    ----------
    D : dict
        LeaderFileImportantImagingParametersRecordALOS dict
    """

    logging.info("===In get_alos_palsar_attitude...")

    A = []
    A.append(D['1st Pitch (degrees)'][2][0])
    A.append(D['1st Roll (degrees)'][2][0])
    A.append(D['1st Yaw (degrees)'][2][0])
    A.append(D['1st Pitch rate (degrees/sec)'][2][0])
    A.append(D['1st Roll rate (degrees/sec)'][2][0])
    A.append(D['1st Yaw rate (degrees/sec)'][2][0])
    A = A + D['2nd, 3rd, ... attitude data points'][2]

    nAttitudeDataPoints = D['Number of attitude data points="bb22", "bb62"'][2][0]
    A = np.array(A)
    A = np.reshape(A, (nAttitudeDataPoints, 6))

    logging.info("===Out get_alos_palsar_attitude.")

    return A


def read_alos_palsar_ldr_iip(ledfile, verbose=False):
    r"""Read important imaging parameters from leader file

    Read important imaging parameters from leader file.

    For example, ``LED-ALPSRP050500980-H1.0__A``

    Parameters
    ----------
    ledfile : str
        Leader file path string.
    """

    logging.info("===In read_alos_palsar_ldr_iip...")
    raise(ImportError('Not opened yet!'))


def read_alos_palsar_raw(rawfile, ledfile, sl=1, el=-1, rmbp=False):
    r"""read ALOS PALSAR raw data

    read ALOS PALSAR raw data pulse from line :attr:`sl` to line :attr:`el`. This function call
    function :func:`read_ceos_sar_raw` firstly and do some post-process.


    Parameters
    ----------
    rawfile : str
        ALOS PALSAR raw data file path string, ---> ``IMG*.0__A``, e.g. ``IMG-HH-ALPSRP050500980-H1.0__A``
    ledfile : str
        ALOS PALSAR leader file path string, ---> ``LED*.0__A``, e.g. ``LED-ALPSRP050500980-H1.0__A``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    Sr : 2d-array
       SAR raw signal data matrix.

    Pd : dict
       Parameter dictionary.

    """

    logging.info("===In read_alos_palsar_raw...")

    raise(ImportError('Not opened yet!'))



def read_alos_palsar_slc(filepath, sl=1, el=-1, rmbp=False):
    r"""read ALOS PALSAR SLC data

    read ALOS PALSAR SLC data pulse from line :attr:`sl` to line :attr:`el`. This function call
    function :func:`read_ceos_sar_slc` firstly.


    Parameters
    ----------
    filepath : str
        ALOS PALSAR raw data file path string, for ALOSPALSAR --> ``IMG*.1__A``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array
       SAR SLC data matrix.

    """

    logging.info("===In read_alos_palsar_slc...")

    raise(ImportError('Not opened yet!'))




#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : simulation_freq_domain.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import math
import torch as th
import torchbox as tb
import torchsar as ts
import matplotlib.pyplot as plt


def dsm2sar_fd(g, pdict, mode='StationaryTarget2D', ftshift=True, device='cpu'):
    """Frequency-domain simulation

    SAR raw data simulation by frequency-domain method.

    Parameters
    ----------
    g : tensor
        The digital scene matrix.
    pdict : dict
        The SAR platform parameters.
    mode : str, optional
        Simulation mode.
    ftshift : bool, optional
        Shift zero-frequency to center?
    device : str, optional
        Specifies which device to be used.

    Returns
    -------
    tensor
        Simulated SAR raw data.
    """
    Vg = pdict['Vg']
    Vr = pdict['Vr']
    Ar = pdict['Ar']
    SC = th.tensor(pdict['SceneCenter'], device=device)

    Kr, Tp = pdict['Kr'], pdict['Tp']
    Na, Nr = pdict['EchoSize']
    if pdict['GeometryMode'] == 'SceneGeometry':
        Rc = pdict['Rsc']
        R0 = pdict['Rs0']
    if pdict['GeometryMode'] == 'BeamGeometry':
        Rc = pdict['Rbc']
        R0 = pdict['Rb0']
    Wl, Fc, La = pdict['Wl'], pdict['Fc'], pdict['La']

    raise(ImportError('Not opened yet!'))




#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : phase_gradient.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import numpy as np
import torch as th
import torchbox as tb
import torchsar as ts
from tqdm import tqdm
from torchbox.utils.const import *
import matplotlib.pyplot as plt


def spotlight_width(H, BWa, Lsar):
    return 2. * H * np.tan(BWa / 2.) - Lsar / (2. * H)


def phase_correction(x, phi, nsub=None, axis=-2, ftshift=True):

    if th.is_complex(x):
        # N, Na, Nr = x.size(0), x.size(-2), x.size(-1)
        cplxflag = True
    elif x.size(-1) == 2:
        # N, Na, Nr = x.size(0), x.size(-3), x.size(-2)
        x = th.view_as_complex(x)
        cplxflag = False
    else:
        raise TypeError('x is complex and should be in complex or real represent formation!')
    if axis == -2 or axis == 1:  # azimuth
        axis = -2
    if axis == -1 or axis == 2:  # range
        axis = -1

    if x.dim() == 2:
        x = x.unsqueeze(0)
    if phi.dim() == 1:
        phi = phi.unsqueeze(0)
    if phi.size(-1) != 1:
            phi = phi.unsqueeze(-1)
            phi = phi.unsqueeze(-1)

    N, Na, Nr = x.size()
    Nx = x.size(axis)
    pshape = [1, 1, 1]
    pshape[0] = N
    pshape[axis] = Nx
    phi = phi.reshape(pshape)

    if nsub is None:
        nsub = Nx

    # X = tb.fft(SI, axis=axis, shift=ftshift)
    # X = X * np.exp(-1j * phi)
    # SI = tb.ifft(X, axis=axis, shift=ftshift)
    xc = x.clone().detach()
    d = xc.dim()
    for n in range(0, Nx, nsub):
        idx = tb.sl(d, axis, range(n, n + nsub))
        xsub = x[idx]
        phisub = phi[idx]
        Xsub = tb.fft(xsub, axis=axis, shift=ftshift)
        xc[idx] = tb.ifft(Xsub * th.exp(-1j * phisub), axis=axis, shift=ftshift)
    if not cplxflag:
        xc = th.view_as_real(xc)
    return xc


def pgaf_sm_1iter(x, windb=None, est='ML', deg=4, axis=-2, isrmlpe=False, iscrct=False, isplot=False):
    r"""perform phase gradient autofocus 1 iter

    Perform phase gradient autofocus 1 iter as described in [1].

    revised for stripmap SAR

    - [1] C.V. Jakowatz, D.E. Wahl, P.H. Eichel, D.C. Ghiglia, P.A. Thompson,
    {\em Spotlight-mode Synthetic Aperture Radar: A Signal Processing
    Approach.} Springer, 1996.

    - [2] D.E. Wahl, P.H. Eichel, D.C. Ghiglia, C.V. Jakowatz, "Phase
    gradient autofocus-a robust tool for high resolution SAR phase
    correction," IEEE Trans. Aero. & Elec. Sys., vol. 30, no. 3, 1994.

    Args:
        x (Tensor): Complex SAR image :math:`{\bm X}\in {\mathbb C}^{N×N_a×N_r}`, where :math:`N` is the batchsize.
        windb (None or float, optional): Cutoff for window (the default is None, which use the mean as the cutoff.)
        est (str, optional): Estimator, ``'ML'`` for Maximum Likelihood estimation, ``'LUMV'`` for Linear Unbiased Minimum Variance estimation (the default is 'ML')
        deg (int): The degree of polynominal
        axis (int, optional): Autofocus axis.
        isrmlpe (bool, optional): Is remove linear phase error? (the default is False)
        iscrct (bool, optional): Is corrected image? (the default is False)
        isplot (bool, optional): Is plot estimated phase error? (the default is False)

    Returns:
        xc (Tensor): Corrected SAR image :math:`{\bm Y}\in {\mathbb C}^{N×N_a×N_r}`, only if :attr:`iscrct` is ``True``, xc is returned.
        phi (Tensor): Estimated phase error :math:`\phi\in {\mathbb R}^{N×N_a}`.

    Raises:
        TypeError: :attr:`x` is complex and should be in complex or real represent formation!
        ValueError: Not supported estimator! Supported are ``'ML'`` and ``'LUMV'``.

    """
    cmap, showidx = 'gray', -1
    raise(ImportError('Not opened yet!'))


def pgaf_sm(SI, nsar, nsub=None, windb=None, est='ML', deg=4, niter=None, tol=1.e-6, isplot=False, islog=False):
    r"""Phase gradient autofocus for stripmap SAR.

    Phase gradient autofocus for stripmap SAR.

    Args:
        SI (Tensor): Complex SAR image :math:`N_a×N_r`.
        nsar (int): Number of synthetic aperture pixels.
        nsub (int, optional): Number of sub-aperture pixels. (the default is :math:`{\rm min}{N_{sar}, N_a}`)
        windb (None or float, optional): cutoff for window (the default is None, which use the mean as the cutoff.)
        est (str, optional): estimator, ``'ML'`` for Maximum Likelihood estimation, ``'LUMV'`` for Linear Unbiased Minimum Variance estimation (the default is 'ML')
        deg (int): Polynomial degrees (default 4) to fit the error, once fitted, the term of deg=[0,1] will be removed.
            If :attr:`deg` is None or lower than 2, then we do not fit the error with polynominal and not remove the linear trend.
        niter (int, optional): Maximum iters (the default is None, which means using :attr:`tol` for stopping)
        tol (float, optional): Phase error tolerance. (the default is 1.e-6)
        isplot (bool, optional): Plot estimated phase error or corrected image. (the default is False)
        islog (bool, optional): Print log information? (the default is False)

    Returns:
        SI (Tensor): Corrected SAR image :math:`{\bm X}\in {\mathbb C}^{N_a×N_r}`.
        phi (Tensor): Estimated phase error :math:`{\phi}\in {\mathbb R}^{N_a×1}`.

    Raises:
        TypeError: The input is complex and should be in complex or real represent formation!
        ValueError: For stripmap SAR, processing sub aperture should be smaller than synthetic aperture!
    """

    cmap, showidx = 'gray', -1
    if islog:
        print("---In pgaf_sm...")

    if th.is_complex(SI):
        # N, Na, Nr = x.size(0), x.size(-2), x.size(-1)
        cplxflag = True
    elif SI.size(-1) == 2:
        # N, Na, Nr = x.size(0), x.size(-3), x.size(-2)
        SI = th.view_as_complex(SI)
        cplxflag = False
    else:
        raise TypeError('SI is complex and should be in complex or real represent formation!')

    if SI.dim() == 2:
        SI = SI.unsqueeze(0)
        # SI = SI.to(th.complex128)

    N, Na, Nr = SI.size()
    if nsub is None:
        nsub = 2**tb.prevpow2(nsar / 2. + 1)
        nsub = min(nsub, Na)
    if nsub > nsar:
        raise ValueError(
            "For stripmap SAR, processing sub aperture %d should be smaller than synthetic aperture %d!" % (nsub, nsar))

    Noverlap = 0

    if niter is None:
        niter = 100
    if islog:
        print("~~~Do Phase Gradient Autofocus...")
    phi0, k = (1.e10, 0)
    phi = th.zeros((N, Na), device=SI.device)
    phitotal = 0.
    for k in tqdm(range(niter)):
        phi0 = phi.clone().detach()
        for n in range(0, Na, nsub):
            idx = n + nsub + Noverlap
            ep, gp = pgaf_sm_1iter(SI[:, n:idx, :], windb=windb, est=est, deg=deg,
                                   isrmlpe=True, iscrct=False, isplot=isplot)
            phi[:, n:n + nsub] = ep[:, 0:nsub]

        phitotal = phitotal + phi
        SI = phase_correction(SI, phi, nsub, axis=-2, ftshift=True)

        if isplot:
            XX = th.abs(SI[showidx])
            XX = tb.mapping(XX)
            plt.figure()
            plt.imshow(XX.numpy(), cmap=cmap)
            plt.show()

        if (th.mean(th.abs(phi - phi0)) <= tol):
            break

    if not cplxflag:
        SI = th.view_as_real(SI)
    
    if islog:
        print("~~~Done.")
        print("---Out pgaf_sm.")

    return SI, phitotal


if __name__ == '__main__':

    datafile = 'data/sarimg/ALPSRP020160970_Vr7180_R3_FocusedEach_5x256x256_AutoFocusPolyPhiMinEntropy_Epoch1000.h5'
    SI = tb.loadmat(datafile)['SI']

    SI0 = SI
    # sa, ea, sr, er = 3500, 3500 + 2048, 5500, 5500 + 2048
    sa, ea, sr, er = 4600, 4600 + 1024, 5000, 5000 + 1024
    # sa, ea, sr, er = 3000, 3000 + 512, 5000, 5000 + 512
    # sa, ea, sr, er = 3200, 3200 + 256, 5000, 5000 + 256

    SI = SI[sa:ea, sr:er, 0] + 1j * SI[sa:ea, sr:er, 1]

    SI0 = SI0[sa:ea, sr:er, 0] + 1j * SI0[sa:ea, sr:er, 1]

    deg = 7
    # deg = None
    est = 'ML'
    est = 'LUMV'

    niter = 40

    SI = th.from_numpy(SI)
    SI0 = th.from_numpy(SI0)
    # SI = SI.to('cuda:0')
    SI, pa = ts.pgaf_sm(SI, 6785, nsub=None, windb=None, est=est, deg=deg, niter=niter, tol=1.e-6, isplot=False, islog=False)

    print(pa.shape)
    plt.figure()
    plt.plot(pa[-1])
    plt.grid()
    plt.xlabel('Aperture time (samples)')
    plt.ylabel('Phase (rad)')
    plt.title('Estimated phase error (polynomial degree ' + str(deg) + ')')
    plt.show()

    SI0 = th.abs(SI0)
    SI = th.abs(SI)
    print("ENT:", tb.entropy(SI0))
    print("ENT:", tb.entropy(SI))
    print("MSE", th.sum(SI0 - SI))

    print(SI.shape, SI0.shape)
    SI = tb.mapping(SI)
    SI0 = tb.mapping(SI0)
    print(SI.shape, SI0.shape)

    plt.figure()
    plt.subplot(121)
    plt.imshow(SI0, cmap='gray')
    plt.subplot(122)
    plt.imshow(SI[-1], cmap='gray')
    plt.show()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : minimum_entropy.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import print_function
import torch as th
import torchbox as tb
import torchsar as ts
from tqdm import tqdm


def __entropy(X):

    if th.is_complex(X):  # N-Na-Nr
        X = (X * X.conj()).real
    elif X.size(-1) == 2:  # N-Na-Nr-2
        X = th.sum(X.pow(2), axis=-1)

    logfunc, axis = th.log2, (1, 2)
    P = th.sum(X, axis=axis, keepdims=True)
    p = X / (P + tb.EPS)
    S = -th.sum(p * logfunc(p + tb.EPS), axis)
    return th.mean(S)


def meaf_ssa_sm(x, niter=10, delta=None, toli=1e-2, tolo=1e-2, ftshift=True, islog=False):
    r"""Stage-by-Stage minimum entropy

    [1] Morrison Jr, Robert Lee; Autofocus, Entropy-based (2002): Entropy-based autofocus for synthetic aperture radar.

    Parameters
    ----------
    x : Tensor
       Corrupt complex SAR image. N-Na-Nr(complex) or N-Na-Nr-2(real)
    niter : int, optional
       The number of iteration (the default is 10)
    delta : {float or None}, optional
       The change step (the default is None (i.e. PI))
    toli : int, optional
       Tolerance error for inner loop (the default is 1e-2)
    tolo : int, optional
       Tolerance error for outer loop (the default is 1e-2)
    ftshift : bool, optional
       Shift the zero frequency to center? (the default is True)
    islog : bool, optional
       Print log information? (the default is False)
    """

    if delta is None:
        delta = tb.PI

    if th.is_complex(x):
        x = th.view_as_real(x)
        cplxflag = True
    elif x.size(-1) == 2:
        cplxflag = False
    else:
        raise TypeError('x is complex and should be in complex or real represent formation!')

    d = x.dim()
    N, Na, Nr = x.size(0), x.size(-3), x.size(-2)

    wshape = [1] * d
    wshape[-3] = Na
    wshape[0] = N

    x = tb.fft(x, cdim=-1, dim=-3, keepcdim=True, shift=ftshift)
    phio = th.zeros(wshape, device=x.device, dtype=x.dtype)
    ephi = th.cat((th.cos(phio), th.sin(phio)), dim=-1)
    # print(wshape, phio.min(), phio.max(), ephi.min(), ephi.max())
    So = __entropy(tb.ifft(tb.ematmul(x, ephi), axis=-3, shift=ftshift))
    i, Soi0, Soo0 = 0, 1e13, 1e13
    # print(i, So, Soi0, Soo0)
    while(i < niter):
        Soo0 = So
        while True:
            Soi0 = So
            # print(i, "===")
            for a in range(Na):
                phi1, phi2 = phio.clone().detach(), phio.clone().detach()
                for n in range(N):
                    phi1[n, a] += delta
                    phi2[n, a] -= delta
                ephi = th.cat((th.cos(phi1), th.sin(phi1)), dim=-1)
                S1 = __entropy(tb.ifft(tb.ematmul(x, ephi), axis=-3, shift=ftshift))
                ephi = th.cat((th.cos(phi2), th.sin(phi2)), dim=-1)
                S2 = __entropy(tb.ifft(tb.ematmul(x, ephi), axis=-3, shift=ftshift))
                if S1 < So:
                    So = S1
                    phio = phi1.clone().detach()
                elif S2 < So:
                    So = S2
                    phio = phi2.clone().detach()
                # print(Soi0, So, S1, S2, abs(So - Soi0), i)
            if abs(So - Soi0) < toli:
                break
        if abs(So - Soo0) > tolo:
            i += 1
            delta /= 2.
        else:
            break
        # print(i, delta, abs(So - Soo0), tolo, abs(So - Soo0) < tolo, So, Soo0)
        # print(i, delta, abs(So - Soi0), tolo, abs(So - Soi0) < toli, So, Soi0)
    ephi = th.cat((th.cos(phio), th.sin(phio)), dim=-1)
    return tb.ifft(tb.ematmul(x, ephi)), ephi


def meaf_sm(x, phi=None, niter=10, tol=1e-4, eta=0.1, method='N-MEA', selscat=False, isunwrap=True, deg=None, axis=-2, ftshift=True, islog=False):
    r"""Entropy based autofocus

    Minimum-Entropy based Autofocus (MEA)

    Args:
        x (Tensor): complex image with shape :math:`N×N_a×N_r` or :math:`N×N_a×N_r×2`
        phi (Tensor, optional): initial value of :math:`\phi` (the default is None, which means zeros)
        niter (int, optional): number of iterations (the default is 10)
        tol (float, optional): Error tolerance.
        eta (float, optional): Learning rate.
        method (str, optional): method used to update the phase error
            - ``'FP-MEA'`` --> Fix Point
            - ``'CD-MEA'`` --> Coordinate Descent
            - ``'SU-MEA'`` --> Simultaneous Update
            - ``'N-MEA'`` --> Newton, see [2], [1] has problem when used to small image
            - ``'SN-MEA'`` --> Simplified Newton
            - ``'MN-MEA'`` --> Modified Newton
        selscat (bool, optional): Select brighter scatters (default: False).
        isunwrap (bool, optional): Unwrap radian phase (default: True).
        deg (int or None, optional): The degree of polynominal, if None, no fitting.
        axis (int, optional): Compensation axis.
        ftshift (bool, optional): Does shift zero frequency to center?
        islog (bool, optional): Does print log info.


        see [1] Zhang S , Liu Y , Li X . Fast Entropy Minimization Based Autofocusing Technique for ISAR Imaging[J]. IEEE Transactions on Signal Processing, 2015, 63(13):3425-3434.
            [2] Zeng T , Wang R , Li F . SAR Image Autofocus Utilizing Minimum-Entropy Criterion[J]. IEEE Geoence & Remote Sensing Letters, 2013, 10(6):1552-1556.

    Returns:
        (Tensor): Description
    """

    if not th.is_complex(x):
        x = th.view_as_complex(x)

    if selscat:
        x = ts.center_dominant_scatters(x)
        x = ts.window_data(x, win=None)

    if x.dim() == 2:
        x = x.unsqueeze(0)

    if axis == -2 or axis == 1:
        axisc, axiso = -2, -1
    if axis == -1 or axis == 2:
        axisc, axiso = -1, -2

    N, Na, Nr = x.size()

    if islog:
        print("---Do Minimum Entropy Autofocus(%s)..." % method)

    phi = th.tensor([0], device=x.device)
    raise(ImportError('Not opened yet!'))


if __name__ == '__main__':

    import matplotlib.pyplot as plt

    matfile = '/mnt/e/ws/github/psar/psar/examples/imaging/data/ALPSRP020160970_Vr7180_R3.mat'
    SI = tb.loadmat(matfile)['SI']

    SI0 = SI
    # sa, ea, sr, er = 3500, 3500 + 2048, 5500, 5500 + 2048
    sa, ea, sr, er = 4600, 4600 + 1024, 5000, 5000 + 1024
    sa, ea, sr, er = 3000, 3000 + 512, 5000, 5000 + 512
    # sa, ea, sr, er = 3200, 3200 + 256, 5000, 5000 + 256

    SI = SI[sa:ea, sr:er, 0] + 1j * SI[sa:ea, sr:er, 1]

    SI0 = SI0[sa:ea, sr:er, 0] + 1j * SI0[sa:ea, sr:er, 1]

    deg = 7
    method, niter = 'N-MEA', 160
    # method, niter = 'FP-MEA', 400

    SI = th.from_numpy(SI)
    SI0 = th.from_numpy(SI0)
    # SI = SI.to('cuda:0')
    SI, pa = ts.meaf_sm(SI, phi=None, niter=niter, tol=1e-4, eta=0.1, method=method, selscat=False, ftshift=True, islog=True)

    print(pa.shape)
    plt.figure()
    plt.plot(pa[-1])
    plt.grid()
    plt.xlabel('Aperture time (samples)')
    plt.ylabel('Phase (rad)')
    plt.title('Estimated phase error (polynomial degree ' + str(deg) + ')')
    plt.show()

    SI0 = th.abs(SI0)
    SI = th.abs(SI)
    print("ENT:", tb.entropy(SI0))
    print("ENT:", tb.entropy(SI))
    print("MSE", th.sum(SI0 - SI))

    print(SI.shape, SI0.shape)
    SI = tb.mapping(SI)
    SI0 = tb.mapping(SI0)
    print(SI.shape, SI0.shape)

    plt.figure()
    plt.subplot(121)
    plt.imshow(SI0, cmap='gray')
    plt.subplot(122)
    plt.imshow(SI[-1], cmap='gray')
    plt.show()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : complex_image.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import torch as th
import torchsar as ts


class CICISTA(object):
    def __init__(self, niter=10, mu=0.5, k=None, eps=1e-4):

        self.niter = niter
        self.mu = mu
        self.eps = eps
        self.k = k
        self.t = 0

        raise(ImportError('Not opened yet!'))



def mfpc_throwaway(X, No, Nh, axis=0, mffdmod='way1', ftshift=False):
    r"""Throwaway invalid pulse compressed data

    Throwaway invalid pulse compressed data

    Parameters
    ----------
    X : Tensor
        Data after pulse compression.
    No : int
        Output size.
    Nh : int
        Filter size.
    axis : int, optional
        Throwaway dimension. (the default is 0)
    mffdmod : str, optional
        Frequency filter mode. (the default is 'way1')
    ftshift : bool, optional
        Whether to shift frequency (the default is False)
    """





class SarPlat(object):
    r"""SAR platform class.
        sensor
        acquisition
        params
    """

    def name(self):
        ...

    def name(self, value):
        ...

    def sensor(self):
        ...

    def sensor(self, value):
        ...

    def acquisition(self):
        ...

    def acquisition(self, value):
        ...

    def params(self):
        ...

    def params(self, value):
        ...

    def selection(self):
        ...

    def selection(self, value):
        ...

    def printsp(self, verbose=False):




def rda_mftd(Sr, pdict, mfmod='fftconv', iqc=False, rcmc=False, dcem=None, win=None, afa=None, ftshift=False, isplot=False, islog=True):
    r"""Range Doppler Algorithm with time domain matched filters

    Args:
        Sr (Tensor): The SAR complex echo data representated in real format :math:`{\bm S}_r \in \mathbb{R}^{N_a×N_r×2}`.
        pdict (dict): Parameters used in RDA, which are show as follows:
            ``{'Vr', 'R0', 'La', 'Fc', 'Tp', 'Kr',  Fsr', 'Fsa'}``
        mfmod (str, optional): Matched filter mode, supported are:

            - ``'corr1'`` : 1d correlation filter and use standard correlation
            - ``'conv1'`` : 1d convolution filter and use standard convolution
            - ``'fftcorr1'`` : 1d correlation filter and use fft for operating correlation
            - ``'fftconv1'`` : 1d convolution filter and use fft for operating convolution (default)

        iqc (bool, optional): Whether do IQ data correction, see :func:`iq_correct`:

            - I/Q bias removal
            - I/Q gain imbalance correction
            - I/Q non-orthogonality correction

        rcmc (bool, int, optional): Range Migration Correction: integer-->kernel size, ``False``-->no rcmc (default: {8})
        dcem (str, optional): Dopplor centroid frequency estimation method (the default is None, does not estimate)
            - ``'abdce_wda'`` :
        win (list, tuple or None, optional): the window function for matched filter of azimuth and range. If None, no window is added (default), e.g. ['kaiser 12', 'hanning'], this will add kaiser window and hanning window in azimuth and range respectively.
        afa (str, optional): Dopplor rate estimation (autofocus) method (the default is None, does not do autofocusing)
        ftshift (bool, optional): Whether to shift zeros frequency to center when use fft, ifft, fftfreq (the default is ``False``)
        isplot (bool, optional): Plot part of processing result, such as DCE result (default: ``False``)
        islog (bool, optional): Display processing info (default: ``True``)

    Returns:
        Tensor: Focused complex image :math:`{\bm S}_i \in \mathbb{C}^{N_a×N_r}`.
    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : gain_compensation.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import torch as th
from torchbox.utils.const import *


def vga_gain_compensation(S, V, mod='linear', fact=1.0):
    r"""vga gain compensation

    vga gain compensation

    .. math::
       \begin{aligned}
       {\bm F} &= (λ 10^{{\bm V}/20})\\
       {\bm S}_{c} &= {\bm F} \odot {\bm S}
       \end{aligned}

    Parameters
    ----------
    S : torch tensor
        :attr:`S` is an :math:`N_a×N_r×2` array, where, :math:`S[:,:,0]` is the I signal
        and :math:`S[:,:,1]` is the Q signal.
    V : torch tensor
        :attr:`S` is an :math:`N_a×N_r` or :math:`N_a×1` VGA gain array, the gain values are in
        ``dB`` unit.
    mod : str, optional
        compensation mode (the default is 'linear')
    fact : number, optional
        fact is the factor :math:`\lambda` (the default is 1.0)

    Returns
    -------
    torch tensor
        compensated signal, :math:`N_a×N_r×2` array.
    """

    Na, Nr, _ = S.shape
    S = S.to(th.float32)
    V = th.tensor(V).to(th.float32)[0:Na, ...]

    linear_gain_factor = fact * 10.**(V / 20.0)

    if V.dim() == 3:
        S = S * linear_gain_factor
    elif V.dim() == 2 and V.shape[1] + V.shape[0] > 2:
        S[:, :, 0] = S[:, :, 0] * linear_gain_factor
        S[:, :, 1] = S[:, :, 1] * linear_gain_factor
    else:
        # A = np.repeat(linear_gain_factor, Nr*2).reshape(Na, Nr, 2)
        # S = A * S

        for n in range(Na):
            S[n, ...] = linear_gain_factor[n] * S[n, ...]

    return S

# def vga_gain_compensation(S, V, mod='linear', fact=1.0):
#     r"""vga gain compensation

#     vga gain compensation

#     .. math::
#        \begin{aligned}
#        {\bm F} &= (λ 10^{{\bm V}/20})\\
#        {\bm S}_{c} &= {\bm F} \odot {\bm S}
#        \end{aligned}

#     Parameters
#     ----------
#     S : array
#         :attr:`S` is an :math:`N_a×N_r×2` array, where, :math:`S[:,:,0]` is the I signal
#         and :math:`S[:,:,1]` is the Q signal.
#     V : array
#         :attr:`S` is an :math:`N_a×N_r` or :math:`N_a×1` VGA gain array, the gain values are in
#         ``dB`` unit.
#     mod : str, optional
#         compensation mode (the default is 'linear')
#     fact : int, optional
#         fact is the factor :math:`\lambda` (the default is 1.0)

#     Returns
#     -------
#     array
#         compensated signal, :math:`N_a×N_r×2` array.
#     """

#     logging.info("===In vga_gain_compensation...")

#     Na, Nr, _ = S.shape
#     if type(S) is not th.Tensor:
#         S = th.from_numpy(S)
#     S = S.to(th.float)
#     V = th.from_numpy(V)[0:Na].reshape(Na, 1)

#     linear_gain_factor = (fact * 10.**(V / 20.0))

#     if np.ndim(V) == 3:
#         S = S * linear_gain_factor
#     elif np.ndim(V) == 2 and V.shape[1] + V.shape[0] > 2:
#         S[:, :, 0] = S[:, :, 0] * linear_gain_factor
#         S[:, :, 1] = S[:, :, 1] * linear_gain_factor
#     else:
#         # A = np.repeat(linear_gain_factor, Nr*2).reshape(Na, Nr, 2)
#         # S = A * S

#         for n in range(Na):
#             S[n, :, :] = linear_gain_factor[n] * S[n, :, :]

#     logging.info("===Out vga_gain_compensation.")

#     return S


from __future__ import absolute_import

from .utils import getnumber, splitfmt
from .record import readrcd, readrcd1item, printrcd
from .ceos import decfmtfceos, read_ceos_sar_raw, read_ceos_sar_slc, SarDataFileFileDescriptorRecordCEOS, SarDataFileSignalDataRecordCEOS, SarDataFileProcessedDataRecordCEOS
from .ers import read_ers_sar_raw, read_ers_sar_slc, SarDataFileFileDescriptorRecordERS, SarDataFileSignalDataRecordERS, SarDataFileProcessedDataRecordERS
from .radarsat import read_radarsat_sar_raw, SarDataFileFileDescriptorRecordRADARSAT, SarDataFileSignalDataRecordRADARSAT
from .uavsar import read_uavsar_csm, read_uavsar_mlc



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : fourier_domain_optim.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import print_function
import torch as th
import torchbox as tb


def __fdnorm(X, p):

    X = th.sum(X.abs(), axis=-1).sqrt()
    # if th.is_complex(X):  # N-Na-Nr
    #     X = X.abs()
    # elif X.size(-1) == 2:  # N-Na-Nr-2
    #     X = th.sum(X.pow(2), axis=-1).sqrt()
    print(th.sum(X), "]]]]")
    C = th.sum(th.sum(X, axis=-2).pow(p), axis=-1)
    print(C.shape, C)
    return th.mean(C)


def af_ffo_sm(x, p=2, niter=10, delta=None, toli=1e-2, tolo=1e-2, ftshift=True, islog=False):
    """Stage-by-Stage minimum entropy

    [1] Morrison Jr, Robert Lee; Autofocus, Entropy-based (2002): Entropy-based autofocus for synthetic aperture radar.

    Parameters
    ----------
    x : Tensor
       Corrupt complex SAR image. N-Na-Nr(complex) or N-Na-Nr-2(real)
    niter : int, optional
       The number of iteration (the default is 10)
    delta : {float or None}, optional
       The change step (the default is None (i.e. PI))
    toli : int, optional
       Tolerance error for inner loop (the default is 1e-2)
    tolo : int, optional
       Tolerance error for outer loop (the default is 1e-2)
    ftshift : bool, optional
       Shift the zero frequency to center? (the default is True)
    islog : bool, optional
       Print log information? (the default is False)
    """

    if delta is None:
        delta = tb.PI

    if th.is_complex(x):
        x = th.view_as_real(x)
        cplxflag = True
    elif x.size(-1) == 2:
        cplxflag = False
    else:
        raise TypeError('x is complex and should be in complex or real represent formation!')

    d = x.dim()
    N, Na, Nr = x.size(0), x.size(-3), x.size(-2)

    wshape = [1] * d
    wshape[-3] = Na

    X = tb.fft(x, cdim=-1, dim=-3, keepcdim=True, shift=ftshift)
    phio = th.zeros(wshape, device=x.device, dtype=x.dtype)
    ephi = th.cat((th.cos(phio), th.sin(phio)), dim=-1)
    # print(wshape, phio.min(), phio.max(), ephi.min(), ephi.max())
    So = __fdnorm(tb.ematmul(X, ephi), p)
    ii, io, Soi0, Soo0 = 0, 0, 1e13, 1e13
    print(ii, So, Soi0, Soo0)
    while(ii < niter):
        Soo0 = So
        while True:
            Soi0 = So
            print(ii, "===", d)
            for a in range(Na):
                phi1, phi2 = phio.clone().detach(), phio.clone().detach()
                for n in range(N):
                    print(n, N, Na, a, delta)
                    phi1[n, a] += delta
                    phi2[n, a] -= delta
                ephi1 = th.cat((th.cos(phi1), -th.sin(phi1)), dim=-1)
                S1 = __fdnorm(tb.ematmul(X, ephi1), p)
                ephi2 = th.cat((th.cos(phi2), -th.sin(phi2)), dim=-1)
                S2 = __fdnorm(tb.ematmul(X, ephi2), p)
                print(phi1.shape, phi2.shape, ephi1.shape, ephi2.shape, "]]]")
                print(N, S1, S2, So, S1 - S2, th.sum(ephi1-ephi2), th.sum(phi1-phi2))
                if S1 > So:
                    So = S1
                    phio = phi1.clone().detach()
                    print("111")
                elif S2 > So:
                    print("222")
                    So = S2
                    phio = phi2.clone().detach()
                print(Soi0, So, S1, S2, abs(So - Soi0), io, "+++")
            if abs(So - Soi0) < toli and io > niter:
                break
            io += 1
        print(ii, delta, abs(So - Soo0), tolo, abs(So - Soo0) < tolo, So, Soo0, "So")
        print(ii, delta, abs(So - Soi0), tolo, abs(So - Soi0) < toli, So, Soi0, "So")
        if abs(So - Soo0) > tolo:
            ii += 1
            delta /= 2.
        else:
            break
    ephi = th.cat((th.cos(phio), th.sin(phio)), dim=-1)
    return tb.ifft(tb.ematmul(X, ephi)), ephi


def af_ffo(g, w=None, p=2, niter=10, eta=0.5, tol=1e-2, ftshift=False):

    if p < 1:
        raise ValueError('p should be larger than 1!')

    G = tb.fft(g, axis=-3, shift=ftshift)
    Na, Nr = G.size(-3), G.size(-2)
    d = G.dim()
    wshape = [1] * d
    wshape[-3] = Na

    if w is None:
        w = th.ones(wshape, device=G.device, dtype=G.dtype)

    if eta is None:
        eta = tb.PI / 2.

    phi0, i = 1e3, 0
    phi = th.zeros(wshape, device=G.device, dtype=G.dtype)
    diff = th.ones(wshape, device=G.device, dtype=G.dtype)
    # print(i, niter, (phi - phi0).sum(), tol)
    # print(w.shape)
    while(i < niter and (phi - phi0).abs().sum() > tol):
       phi0 = phi
       while(diff.abs().sum() > tol):
          Gabs = th.sqrt(G[..., 0] ** 2 + G[..., 1] ** 2).unsqueeze(-1)
          # print(Gabs.shape, w.shape)
          gamman = th.sum(w * Gabs, axis=-3, keepdim=True)
          print(gamman.min(), gamman.max(), "===")
          # print(gamman.shape, w.shape, Gabs.shape)
          Dphi = (w * p * th.sum((gamman ** (p - 1)) *  Gabs, axis=-2, keepdim=True) / 2.) / \
            (-w * w * p * (p - 1) * th.sum((gamman ** (p - 2)) * (Gabs ** 2), axis=-2, keepdim=True) / 4.)

          # print(Dphi.shape)
          diff = eta * Dphi
          phi = phi - diff
          epa = th.cat((th.cos(phi), th.sin(phi)), axis=-1)
          G = tb.ematmul(G, epa)
          print(Dphi.min(), Dphi.max())
          print(diff.min(), diff.max())
          print(phi.min(), phi.max())
       i += 1
       eta /= 2.
       print(i, eta)

    g = tb.ifft(G, axis=-3, shift=ftshift)
    return g



def decfmtfceos(F, n, x, b, e='<'):
    r"""decode format descriptor of CEOS

    decode format descriptor of CEOS.

    ``'nFx'`` --> decode :attr:`b` x Bytes-by-x Bytes as format :attr:`F` specified,
    and do :attr:`n` times.

    Parameters
    ----------
    F : str
        formation string, ``C`` --> Complex(real, imag), ``F`` --> float, ``B`` --> Bytes, ``A`` --> String, ``I`` --> asicc
    n : int
        amount of elements with formation :attr:`F`
    x : int
        number of Bytes
    b : bytes
        bytes to be decoded
    e : str, optional
        endian, ``'<'`` --> little, ``'>'`` -- > big (the default is '<', which means little endian)

    Returns
    -------
    list
        decoded result list.
    """

def _getdtype_component(fmtrcd, fmtuser):
    ...

def read_ceos_sar_raw(filepath, sl=1, el=-1, rmbp=False):
    r"""read CEOS SAR raw data

    read CEOS SAR raw data pulse from line :attr:`sl` to line :attr:`el`.

    Parameters
    ----------
    filepath : str
        SAR raw data file path string, for ERS1/2 --> ``*.raw`` for RADARSAT --> ``DAT_01.001``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array


    """

def read_ceos_sar_slc(filepath, sl=1, el=-1, rmbp=False):
    r"""read CEOS SAR slc data

    read CEOS SAR slc data from line :attr:`sl` to line :attr:`el`.

    Parameters
    ----------
    filepath : str
        SAR slc data file path string, for ERS1/2 --> ``*.D`` for RADARSAT --> ``DAT_01.001``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array
       Processed SLC SAR data matrix.

    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : sar_signal.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import torch as th
from torchbox.utils.const import *
from torchbox.dsp.normalsignals import rect


def sar_tran(t, Tp, Kr, Fc, A=1.):

    return A * rect(t / Tp) * th.exp(2j * PI * Fc * t + 1j * PI * Kr * t**2)


def sar_recv(t, tau, Tp, Kr, Fc, A=1.):

    t = t - tau
    return A * rect(t / Tp) * th.exp(2j * PI * Fc * t + 1j * PI * Kr * t**2)


if __name__ == '__main__':

    import matplotlib.pyplot as plt
    from torch.fft import fft, fftshift

    Kr = 40e+12
    Tp = 2.5e-06
    Br = abs(Kr) * Tp

    alpha = 1.24588  # 1.1-1.4
    Fsr = alpha * Br
    # Fc = 5.3e9
    Fc = 0.

    Tsr = 1.2 * Tp
    Nsr = int(Fsr * Tsr)
    t = th.linspace(-Tsr / 2., Tsr / 2, Nsr)
    f = th.linspace(-Fsr / 2., Fsr / 2, Nsr)

    St = sar_tran(t, Tp, Kr, Fc)

    Yt = fftshift(fft(fftshift(St, dim=0), dim=0), dim=0)

    plt.figure(1)
    plt.subplot(221)
    plt.plot(t * 1e6, th.real(St))
    plt.plot(t * 1e6, th.abs(St))
    plt.grid()
    plt.legend({'Real part', 'Amplitude'})
    plt.title('Matched filter')
    plt.xlabel('Time/μs')
    plt.ylabel('Amplitude')
    plt.subplot(222)
    plt.plot(t * 1e6, th.angle(St))
    plt.grid()
    plt.subplot(223)
    plt.plot(f, th.abs(Yt))
    plt.grid()
    plt.subplot(224)
    plt.plot(f, th.angle(Yt))
    plt.grid()
    plt.show()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : doppler_centroid_estimation.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import torch as th
import numpy as np
from torchbox.utils.const import *
from torchbox.dsp.ffts import fft, ifft, fftfreq
from torchbox.dsp.polynomialfit import polyfit, polyval
import matplotlib.pyplot as plt


def bdce_sf(Sr, Fsa, Fsr, rdflag=False, Nroff=0, isplot=False):
    r"""Baseband doppler centroid estimation by spectrum fitting

    Baseband doppler centroid estimation by spectrum fitting.

    Parameters
    ----------
    Sr : numpy array
        SAR signal :math:`N_a×N_r` in range-doppler domain (range frequency domain).
    Fsa : float
        Sampling rate in azimuth
    rdflag : bool
        Specifies whether the input SAR signal is in range-doppler domain. If not, :func:`dce_sf` excutes FFT in range direction.
    isplot : bool
        Whether to plot the estimated results.(Default: False)
    """
    Na, Nr = Sr.shape

    raise(ImportError('Not opened yet!'))



def bdce_api(Sr, Fsa, isplot=False):
    r"""Baseband doppler centroid estimation by average phase increment

    Baseband doppler centroid estimation by average phase increment.

    Parameters
    ----------
    Sr : numpy array
        SAR raw data or range compressed data
    Fsa : float
        Sampling rate in azimuth
    isplot : bool
        Whether to plot the estimated results.(Default: False)
    """
    Na, Nr = Sr.shape
    raise(ImportError('Not opened yet!'))



def bdce_madsen(Sr, Fsa, isplot=False):
    r"""Baseband doppler centroid estimation by madsen

    Baseband doppler centroid estimation bymadsen.

    Parameters
    ----------
    Sr : numpy array
        SAR raw data or range compressed data
    Fsa : float
        Sampling rate in azimuth
    isplot : bool
        Whether to plot the estimated results.(Default: False)
    """

    raise(ImportError('Not opened yet!'))



def abdce_wda_ori(Sr, Fsa, Fsr, Fc, rate=0.9, isplot=False, islog=False):
    r"""Absolute and baseband doppler centroid estimation by wavelength diversity algorithm

    Absolute and baseband doppler centroid estimation by Wavelength Diversity Algorithm (WDA).

    <<合成孔径雷达成像_算法与实现>> p350.

    Parameters
    ----------
    Sr : numpy array
        SAR signal :math:`N_a×N_r` in range frequency domain.
    Fsa : float
        Sampling rate in azimuth.
    Fsr : float
        Sampling rate in range.
    """
    if islog:
        print("---In abdce_wda_ori...")

    Na, Nr = Sr.shape

    acccv = th.sum(th.conj(Sr[1:, :]) * Sr[0:-1, :], 0)
    phi_acccv = th.angle(acccv)

    raise(ImportError('Not opened yet!'))



def abdce_wda_opt(Sr, Fsr, Fsa, Fc, ncpb=None, tr=None, isfftr=False, isplot=False, islog=False):
    """Absolute and baseband doppler centroid estimation by wavelength diversity algorithm

    Absolute and baseband doppler centroid estimation by Wavelength Diversity Algorithm (WDA).

    <<合成孔径雷达成像_算法与实现>> p350.

    Parameters
    ----------
    Sr : 2d-tensor
        SAR signal :math:`N_a×N_r` in range frequency domain.
    Fsr : float
        Sampling rate in range, unit Hz.
    Fsa : float
        Sampling rate in azimuth, unit Hz.
    Fc : float
        Carrier frequency, unit Hz.
    ncpb : tuple or list, optional
        Number of cells per block, so we have blocks (int(Na/ncpb[0])) × (int(Nr/ncpb[1]))
        (the default is [Na, Nr], which means all).
    tr : 1d-tensor, optional
        Time in range (the default is None, which linspace(0, Nr, Nr)).
    isplot : bool, optional
        Whether to plot the estimation results (the default is False).
    isfftr : bool, optional
        Whether to do FFT in range (the default is False).

    Returns
    -------
    fadc : 2d-tensor
        Absolute doppler centroid frequency, which has the size specified by :attr:`ncpb`.
    fbdc : 2d-tensor
        Baseband doppler centroid frequency, which has the size specified by :attr:`ncpb`.
    Ma : 2d-tensor
        Doppler ambiguity number, which has the size specified by :attr:`ncpb`.
    """

    if islog:
        print("---In abdce_wda_opt...")

    Na, Nr = Sr.shape
    if tr is None:
        tr = th.linspace(0, Nr, Nr)

    ftshift = True

    raise(ImportError('Not opened yet!'))



def fullfadc(fdc, shape):
    nblks = fdc.shape
    Na, Nr = shape

    NBa = nblks[0]
    NBr = nblks[1]
    Na1b = int(np.uint(Na / NBa))
    Nr1b = int(np.uint(Nr / NBr))

    fc = th.zeros((Na, Nr))

    for a in range(NBa):
        for r in range(NBr):
            fc[int(a * Na1b):min(int((a + 1) * Na1b), Na), int(r * Nr1b):min(int((r + 1) * Nr1b), Nr)] = fdc[a, r]
    return fc


if __name__ == "__main__":

    import matplotlib.pyplot as plt
    import torchsar

    datafile = '/mnt/d/DataSets/sar/ALOSPALSAR/mat/ALPSRP050500980-L1.0/ALOS_PALSAR_RAW=IMG-HH-ALPSRP050500980-H1(sl=1el=35345).mat'
    # datafile = '/mnt/d/DataSets/sar/ERS/mat/ERS2_SAR_RAW=E2_02558_STD_L0_F327(sl=1el=28682).mat'

    sardata, sarplat = torchsar.sarread(datafile)

    Fsa = sarplat.params['Fsa']
    Fsr = sarplat.params['Fsr']
    fr = sarplat.params['fr']
    Kr = sarplat.params['Kr']
    Fc = sarplat.sensor['Fc']
    Sr = sardata.rawdata[:, :, 0] + 1j * sardata.rawdata[:, :, 1]

    Sr = Sr[0:1024, 0:2048]
    # Sr = Sr[1024:4096, 0:2048]
    fr = th.linspace(-Fsr / 2.0, Fsr / 2.0, Sr.shape[1])

    # Sr = fftshift(fft(fftshift(Sr, axes=1), axis=1), axes=1)

    # Sr = torchsar.range_matched_filtering(Sr, fr, Kr)

    # Sr = ifftshift(ifft(ifftshift(Sr, axes=1), axis=1), axes=1)

    # aa = torchsar.doppler_center_estimation(Sr, Fsa)
    # print(aa)

    # Sr = Sr[512:-512, 512:-512]
    # Sr = Sr[:, 512:512+1024]
    # Sr = Sr[0:512, 0:512]

    # Sr = fftshift(fft(fftshift(Sr, axes=1), axis=1), axes=1)
    # Sr = fftshift(fft(Sr, axis=1), axes=1)
    # Sr = fft(Sr, axis=1)

    print(Sr.shape)

    accc(Sr, isplot=True)

    _, dc_coef = bdce_sf(Sr, Fsa, rdflag=False, isplot=True)
    print(dc_coef)
    bdce_api(Sr, Fsa, isplot=True)

    fadc = abdce_wda_ori(Sr, Fsa, Fsr, Fc)

    print(fadc)



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : range_doppler_mftd.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import torch as th
import numpy as np
import torchbox as tb
import torchsar as ts
import matplotlib.pyplot as plt


def rda_mftd(Sr, pdict, mfmod='fftconv', iqc=False, rcmc=False, dcem=None, win=None, afa=None, ftshift=False, isplot=False, islog=True):
    r"""Range Doppler Algorithm with time domain matched filters

    Args:
        Sr (Tensor): The SAR complex echo data representated in real format :math:`{\bm S}_r \in \mathbb{R}^{N_a×N_r×2}`.
        pdict (dict): Parameters used in RDA, which are show as follows:
            ``{'Vr', 'R0', 'La', 'Fc', 'Tp', 'Kr',  Fsr', 'Fsa'}``
        mfmod (str, optional): Matched filter mode, supported are:

            - ``'corr1'`` : 1d correlation filter and use standard correlation
            - ``'conv1'`` : 1d convolution filter and use standard convolution
            - ``'fftcorr1'`` : 1d correlation filter and use fft for operating correlation
            - ``'fftconv1'`` : 1d convolution filter and use fft for operating convolution (default)

        iqc (bool, optional): Whether do IQ data correction, see :func:`iq_correct`:

            - I/Q bias removal
            - I/Q gain imbalance correction
            - I/Q non-orthogonality correction

        rcmc (bool, int, optional): Range Migration Correction: integer-->kernel size, ``False``-->no rcmc (default: {8})
        dcem (str, optional): Dopplor centroid frequency estimation method (the default is None, does not estimate)
            - ``'abdce_wda'`` :
        win (list, tuple or None, optional): the window function for matched filter of azimuth and range. If None, no window is added (default), e.g. ['kaiser 12', 'hanning'], this will add kaiser window and hanning window in azimuth and range respectively.
        afa (str, optional): Dopplor rate estimation (autofocus) method (the default is None, does not do autofocusing)
        ftshift (bool, optional): Whether to shift zeros frequency to center when use fft, ifft, fftfreq (the default is ``False``)
        isplot (bool, optional): Plot part of processing result, such as DCE result (default: ``False``)
        islog (bool, optional): Display processing info (default: ``True``)

    Returns:
        Tensor: Focused complex image :math:`{\bm S}_i \in \mathbb{C}^{N_a×N_r}`.
    """

    device = Sr.device
    Na, Nr, _ = Sr.shape

    Vr = pdict['Vr']
    Rnear = pdict['Rnear']
    La = pdict['La']
    Fc = pdict['Fc']
    Tp = pdict['Tp']
    Kr = pdict['Kr']
    Fsr = pdict['Fsr']
    Fsa = pdict['Fsa']
    fdc = pdict['fdc']
    Nsar = pdict['Nsar']
    if type(fdc) is float or type(fdc) is int or type(fdc) is np.float64 or type(fdc) is np.float32:
        fdc = th.tensor([fdc] * Nr, device=device)

    Wl = tb.C / Fc
    Rfar = ts.min_slant_range(Rnear, Fsr, Nr)

    tnear = 2. * Rnear / tb.C
    tfar = tnear + Nr / Fsr
    # tfar = tnear + Nr / Fsr + Tp
    tr = th.linspace(tnear, tfar, Nr).reshape(1, Nr).to(device)

    # tr = pdict['tr'].to(device)
    # Rnear = pdict['Rnear']

    fa = tb.fftfreq(Na, Fsa, norm=False, shift=ftshift).to(device)

    if mfmod in ['fftconv1', 'FFTCONV1', 'fftconv', 'FFTCONV']:
        mffunc, mftdmod = tb.fftconv1, 'conv'
    if mfmod in ['fftcorr1', 'FFTCORR1', 'fftcorr', 'FFTCORR']:
        mffunc, mftdmod = tb.fftcorr1, 'corr'
    if mfmod in ['conv1', 'CONV1', 'conv', 'CONV']:
        mffunc, mftdmod = ts.conv1, 'conv'
    if mfmod in ['corr1', 'CORR1', 'corr', 'CORR']:
        mffunc, mftdmod = ts.corr1, 'corr'

    if iqc:
        Sr, Flag = ts.iq_correct(Sr)
        if islog:
            print(Flag)
    Sr = Sr[..., 0] + 1j * Sr[..., 1]
    # Sr = th.view_as_complex(Sr)

    raise(ImportError('Not opened yet!'))



def read_uavsar_csm(filepath, sl=1, el=-1, rmbp=False):
    r"""read UAVSAR raw data

    read UAVSAR SAR raw data pulse from line :attr:`sl` to line :attr:`el`. This function call
    function :func:`read_ceos_sar_raw` firstly and do some post-process.


    Parameters
    ----------
    filepath : str
        UAVSAR raw data file path string, for UAVSAR --> ``*.raw``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array
       SAR raw signal data matrix.

    """

def read_uavsar_mlc(filepath, dshape=None, dtype='complex'):
    r"""read UAVSAR MLC data

    read UAVSAR SAR MLC data.


    Parameters
    ----------
    filepath : str
        UAVSAR MLC data file path string, for UAVSAR --> ``*.mlc``
    dshape : tuple, optional
        MLC data shape (:math:`N_a\times N_r`), where, :math:`N_a, N_r` equal
        to ``mlc_mag.set_rows``, ``mlc_mag.set_cols`` which can be obtained
        from the ``.ann`` file (the default is 1, which means the first line)
    dshape : str
        MLC data type: ``'complex'`` for complex-valued floating point data,
        ``'real'`` for real-valued floating point data.

    Returns
    -------
    S : 2d-array
       SAR MLC data matrix.

    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : sensor_simulated.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from torchbox.utils.const import *

SENSOR_SIMULATED = {

    'Air1': {
        'Fc': 5.3e9,  # Hz
        'H': 10000,  # height in m
        'V': 150.0,  # velocity in m/s
        'Tp': 2.5e-6,  # Range pulse length in seconds
        'Kr': 20e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 12,  # antenna length (range) in m
        'La': 3.0,  # antenna length (azimuth) in m
        'PRF': 120.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 60.0e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [256, 320],
        'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        # 'As': 2 * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        # 'Aon': 59.3392 * PI / 180.0,  # off-nadir angle |\
        'Aon': 59.33 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Air1',
    },

    'Air2': {
        'Fc': 9.4e9,  # Hz
        'H': 8000,  # height in m
        'V': 200.0,  # velocity in m/s
        'Tp': 1e-6,  # Range pulse length in seconds
        'Kr': 160e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 12,  # antenna length (range) in m
        'La': 3.0,  # antenna length (azimuth) in m
        'PRF': 90.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 96e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [128, 128],
        'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        # 'As': 0.5 * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 58.7 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Air2',
    },

    'Air3': {
        'Fc': 9.4e9,  # Hz
        'H': 8000,  # height in m
        'V': 200.0,  # velocity in m/s
        'Tp': 1e-6,  # Range pulse length in seconds
        'Kr': 160e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 12,  # antenna length (range) in m
        'La': 3,  # antenna length (azimuth) in m
        'PRF': 300.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 192.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [256, 256],
        # 'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'As': 0. * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 67.7 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Air3',
    },

    'Air4': {
        'Fc': 9.4e9,  # Hz
        'H': 8000,  # height in m
        'V': 200.0,  # velocity in m/s
        'Tp': 1e-6,  # Range pulse length in seconds
        'Kr': 80e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 12,  # antenna length (range) in m
        'La': 6,  # antenna length (azimuth) in m
        'PRF': 100.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 96.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [64, 64],
        # 'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'As': 0. * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 67.7 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Air4',
    },

    'Air5': {
        'Fc': 9.6e9,  # Hz
        'H': 8000,  # height in m
        'V': 200.0,  # velocity in m/s
        'Tp':1e-6,  # Range pulse length in seconds
        'Kr': 80e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 4,  # antenna length (range) in m
        'La': 4,  # antenna length (azimuth) in m
        'PRF': 124.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 96.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [128, 128],
        # 'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'As': 0. * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 73.0 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Air5',
    },

    'Air6': {
        'Fc': 9.6e9,  # Hz
        'H': 8000,  # height in m
        'V': 200.0,  # velocity in m/s
        'Tp':1e-6,  # Range pulse length in seconds
        'Kr': 80e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 4,  # antenna length (range) in m
        'La': 4,  # antenna length (azimuth) in m
        'PRF': 62.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 48.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [64, 64],
        # 'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'As': 0. * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 73.0 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Air6',
    },

    'Sim1': {
        'Fc': 5.3e9,  # Hz
        'H': 1000,  # height in m
        'V': 150.0,  # velocity in m/s
        'Tp': 2.5e-6,  # Range pulse length in seconds
        'Kr': 40.e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 2,  # antenna length (range) in m
        'La': 3,  # antenna length (azimuth) in m
        'PRF': 180.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 120.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [512, 512],
        # 'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'As': 0. * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 67.7 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Sim1',
    },

    'Sim2': {
        'Fc': 5.3e9,  # Hz
        'H': 1000,  # height in m
        'V': 150.0,  # velocity in m/s
        'Tp': 2.5e-6,  # Range pulse length in seconds
        'Kr': 40.e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 2,  # antenna length (range) in m
        'La': 3,  # antenna length (azimuth) in m
        'PRF': 180.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 120.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [256, 256],
        # 'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'As': 0. * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 67.7 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Sim2',
    },

    'Sim3': {
        'Fc': 5.3e9,  # Hz
        'H': 10000,  # height in m
        'V': 150.0,  # velocity in m/s
        'Tp': 0.5e-6,  # Range pulse length in seconds
        'Kr': 160e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 2,  # antenna length (range) in m
        'La': 3,  # antenna length (azimuth) in m
        'PRF': 160.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 96.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [128, 128],
        # 'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'As': 0. * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 45. * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Sim3',
    },

    'Sim4': {
        'Fc': 5.3e9,  # Hz
        'H': 1000,  # height in m
        'V': 150.0,  # velocity in m/s
        'Tp': 0.5e-6,  # Range pulse length in seconds
        'Kr': 160.e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 2,  # antenna length (range) in m
        'La': 3,  # antenna length (azimuth) in m
        'PRF': 80.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 48.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [64, 64],
        # 'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'As': 0. * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 45. * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Sim4',
    },

    'Sim5': {
        'Fc': 5.3e9,  # Hz
        'H': 1100,  # height in m
        'V': 150.0,  # velocity in m/s
        'Tp': 0.5e-6,  # Range pulse length in seconds
        'Kr': 80.e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 2,  # antenna length (range) in m
        'La': 5,  # antenna length (azimuth) in m
        'PRF': 100.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 48.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [64, 64],
        # 'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'As': 0. * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 30. * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Sim5',
    },

    'Sim6': {
        'Fc': 5.3e9,  # Hz
        'H': 1100,  # height in m
        'V': 150.0,  # velocity in m/s
        'Tp': 0.5e-6,  # Range pulse length in seconds
        'Kr': 80.e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 2,  # antenna length (range) in m
        'La': 5,  # antenna length (azimuth) in m
        'PRF': 100.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 48.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [32, 32],
        # 'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'As': 0. * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 30. * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Sim6',
    },

}



def convert_ppec(cs, degs=None, mode='fftfreq->2fftfreq'):
    r"""convert polynominal phase error

    Convert polynominal phase error.

    Args:
        cs (Tensor): The polynominal coefficients.
        degs (None, optional): The degrees of polynominal.
        mode (str, optional): ``'fftfreq->2fftfreq'`` or ``'2fftfreq->fftfreq'`` (the default is 'fftfeq->2fftfreq')

    Returns:
        Tensor: Converted phase error tensor.

    Raises:
        ValueError: Not support mode.
    """

def ppeaxis(n, norm=True, shift=True, mode='fftfreq', dtype=th.float32, device='cpu'):
    r"""Return the sample axis

    Return the sample frequencies axis.

    Parameters
    ----------
    n : int
        Number of samples.
    norm : bool
        Normalize the frequencies.
    shift : bool
        Does shift the zero frequency to center.
    mode : str or None
        Frequency sampling mode. ``'fftfreq'`` or ``'freq'`` normalize the frequencies in (-1/2, 1/2).
        ``'2fftfreq'`` or ``'2freq'`` normalize the frequencies in (-1, 1). If ``None``, return ``None``.
    dtype : Tensor
        Data type, default is ``th.float32``.
    device : str
        device string, default is ``'cpu'``.

    Returns
    -------
    torch array
        Frequency array with size :math:`n×1`.
    """

def polype(c, x, deg=None):
    r"""compute phase error with polynominal model

    compute phase error with polynominal model

    Args:
        c (Tensor, optional): The polynominal coefficients matrix :math:`{\bf P} = [{\bf p}_1, {\bf p}_2, \cdots, {\bf p}_N ]^T ∈{\mathbb R}^{N × (M -1)}`,
            where, :math:`{\bf p}_n = [p_2, \cdots, p_M]^T`).
        x (Tensor or None, optional): Axis, such as a vector in range of (-1, 1). If :attr:`x` is None, return :attr:`c` directly.
        deg (tuple or None, optional): The degree of the coefficients, :math:`[p_{\rm min}, p_{\rm_max}]`, default is ``None``, which means ``[2, size(c, 1) + 1]``

    Returns:
        (Tensor): The phase error tensor.

    Raises:
        TypeError: :attr:`c` should not be None!
    """

def dctpe(c, x, deg=None):
    r"""compute phase error with DCT model

    compute phase error with DCT model

    .. math::
        \phi_{e}(h)=\sum_{p=0}^{P} a(p) \cdot d c t(p) \cdot \cos \left[\frac{\pi\left(2 h_{a}+1\right) p}{2 N}\right]

    where, :math:`a(p)=\left{\begin{array}{ll}1 / \sqrt{N} p=0 \ \sqrt{2 / N} p \neq 0\end{array},\right.` and
    :math:`\left{\begin{array}{l}p=0,1,2, \cdots \cdots \cdots, P \ h_{a}=[0,1,2, \cdots \cdots, N-1]\end{array}\right.`

    Args:
        c (Tensor, optional): The DCT coefficients matrix :math:`{\bf P} = [{\bf p}_1, {\bf p}_2, \cdots, {\bf p}_N ]^T ∈{\mathbb R}^{N × (M + 1)}`,
            where, :math:`{\bf p}_n = [p_0, \cdots, p_M]^T`).
        x (Tensor, optional): Axis, such as :math:`(0, Na)`.
        deg (tuple or None, optional): The degree of the coefficients, :math:`[p_{\rm min}, p_{\rm_max}]`, default is ``None``, which means ``[0, size(c, 1)]``

    Returns:
        (Tensor): The phase error tensor.

    Raises:
        TypeError: :attr:`c` should not be None!
    """

def rmlpe(phi, mode='poly', deg=4, ftshift=True):
    r"""remove linear phase error

    Remove linear phase error based on linear fitting, the linear phase error
    will cause image shift.

    Args:
        phi (Tensor): A :math:`N×N_x` phase error tensor with linear trend, where, :math:`N` is the batchsize, :math:`N_x` is the samples numbers in direction x.
        mode (str): model mode, default is polynominal
        deg (int): Polynomial degrees (default 4) to fit the error, once fitted, the term of deg=[0,1] will be removed. If :attr:`deg` is None or lower than 2, then we do not fit the error with polynominal and not remove the linear trend.
        ftshift (bool): Shift zero frequency to center when do fft/ifft/fftfreq? (the default is False)

    Returns:
        phi (tensor): Phase without linear trend.
    """

class PolyPhaseErrorGenerator(object):
    ...

    def __init__(self, carange=None, crrange=None, seed=None):
        """Polynominal phase error generator.

        Args:
            carange (None or tuple or list, optional): List of coefficients range of phase error in azimuth direction.
            crrange (None or tuple or list, optional): List of coefficients range of phase error in range direction.
            seed (None or int, optional): The random seed.
        """

    def mkpec(self, n, seed=None):
        """Make phase error

        Args:
            n (int): The number of phase errors.
            seed (None or int, optional): The random seed.

        Returns:
            TYPE: Description
        """





def iq_correct(Sr):
    r"""IQ Correction performs the I Q data correction

    IQ Correction performs the I Q data correction

    - I/Q bias removal
    - I/Q gain imbalance correction
    - I/Q non-orthogonality correction

    see "Sentinel-1-Level-1-Detailed-Algorithm-Definition".

    Args:
        Sr (Tensor): SAR raw data matrix :math:`{\bm S}_r \in {\mathbb R}^{N_a×N_r×2}`

    Returns:
        Sr (Tensor): Corrected SAR raw data.
        Flag : dict

    """





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015-2130, Zhi Liu.  All rights reserved.

import os
from setuptools import setup
from setuptools import find_packages
from Cython.Build import cythonize
from Cython.Distutils import Extension

pkgname = 'torchsar'
this_dir = os.path.abspath(os.path.dirname(__file__))

def read_requirements(filename):
    return [line.strip() for line in open(filename, encoding="utf-8").read().splitlines()
            if (not line.startswith('#') and len(line)>0)]

def listxfile(listdir=None, exts=None, recursive=True, filelist=[]):
    if listdir is None:
        return filelist
    exts = [exts] if type(exts) is str else exts
    for s in os.listdir(listdir):
        newDir = os.path.join(listdir, s)
        if os.path.isfile(newDir):
            if exts is not None:
                if newDir and(os.path.splitext(newDir)[1] in exts):
                    filelist.append(newDir)
            else:
                filelist.append(newDir)
        else:
            if recursive:
                listxfile(listdir=newDir, exts=exts, recursive=True, filelist=filelist)
    return filelist

version = open(pkgname + '/version.py', encoding="utf-8").read().strip().split('=')[-1].strip().strip('\'').strip('\"')
requirements = read_requirements('requirements.txt')
long_description = open('README.md', encoding="utf-8").read()

modules = listxfile(pkgname, '.py')

py_extensions, c_extensions = [], []
for efile in modules:
    if efile.find('__init__.py') < 0:
        efile = efile[:-3]
        py_extensions.append(Extension(efile.replace('/', '.'), [efile + '.py']))
        c_extensions.append(Extension(efile.replace('/', '.'), [efile + '.c']))

try:
    cythonize(py_extensions, language_level=3)

    setup(name=pkgname,
        version=version,
        description="PyTorch Tools for Synthetic Aperture Radar.",
        long_description=long_description,
        author='Zhi Liu',
        author_email='zhiliu.mind@gmail.com',
        url='https://sar.iridescent.ink/',
        download_url='https://github.com/antsfamily/%s/' % pkgname,
        license='MIT',
        packages=find_packages(),
        install_requires=requirements,
        include_package_data=False,
        keywords=['PyTorch', 'Synthetic Aperture Radar', 'Machine Learning', 'Deep Learning'],
        ext_modules=cythonize(c_extensions, language_level=3)
    )
except:
    setup(name=pkgname,
        version=version,
        description="PyTorch Tools for Synthetic Aperture Radar.",
        long_description=long_description,
        author='Zhi Liu',
        author_email='zhiliu.mind@gmail.com',
        url='https://sar.iridescent.ink/',
        download_url='https://github.com/antsfamily/%s/' % pkgname,
        license='MIT',
        packages=find_packages(),
        install_requires=requirements,
        include_package_data=False,
        keywords=['PyTorch', 'Synthetic Aperture Radar', 'Machine Learning', 'Deep Learning'],
    )


#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : target_background.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import numpy as np
import torch as th
from skimage import filters
from torchbox.utils.const import EPS
import matplotlib.pyplot as plt


def extract_targets(x, region=None, thresh=None, cdim=None, isshow=False):
    r"""Extracts the pixels of targetss

    Parameters
    ----------
    x : tensor or numpy array
        The input image, if it's complex-valued, it's amplitude is used.
    region : list or None, optional
        The region ([top-left, bottom-right]) that contains targets for computing TBR.
        If :obj:`None`, :attr:`region` equals to ``([0, 0], [H, W])``, where :math:`H,W` 
        are the height and width of the image.
    thresh : float or None, optional
        The threshold targets.
    cdim : int or None, optional
        Specifies the complex axis of :attr:`x`. ``None`` --> complex or real
    isshow : bool, optional
        Show pixels of the extracted targets?
    """

    if th.is_complex(x):
        cdim = None
    if cdim is not None:
        x = x.pow(2).sum(cdim).sqrt()
    else:
        x = x.abs()
    x = x.numpy()

    if region is None:
        region = [min(targets[0]), min(targets[1]), max(targets[0]), max(targets[1])]

    x = x[region[0]:region[2], region[1]:region[3]]

    thresh = filters.threshold_otsu(x) if thresh is None else thresh

    idx = np.where(x > thresh)

    if isshow:
        plt.figure
        plt.subplot(121)
        plt.imshow(x)
        y = np.zeros(x.shape)
        y[idx] = 1
        plt.subplot(122)
        plt.imshow(x * y)
        plt.show()
    idx = list(idx)
    idx[0] += region[0]
    idx[1] += region[1]

    return tuple(idx), thresh


def tbr(x, targets, region=None, cdim=None, isshow=False):
    r"""Target-to-Background Ratio (TBR)

    .. math::
        \begin{align}
        {\rm TBR} = 20{\rm log}_{10}\left(\frac{{\rm max}_{i\in{\mathbb T}}(|{\bf X}_i|)}{{(1/N_{\mathbb B})}\Sigma_{j\in \mathbb B}|{\bf X}_j|)}\right)
        \label{equ:TBR}
        \end{align}

    Parameters
    ----------
    x : tensor
        The input image, if it's complex-valued, it's amplitude is used.
    targets : list or tuple
        The targets pixels. ([row1, row2, ...], [col1, col2, ...])
    region : list, tuple or None, optional
        The region ([top-left, bottom-right]) that contains targets for computing TBR.
        If :obj:`None`, :attr:`region` equals to ``([0, 0], [H, W])``, where :math:`H,W` 
        are the height and width of the image.
    cdim : int or None, optional
        Specifies the complex axis of :attr:`x`. ``None`` --> complex or real
    isshow : bool, optional
        Show target mask? (default: False)
    """

    if th.is_complex(x):
        cdim = None
    if cdim is not None:
        x = x.pow(2).sum(cdim).sqrt()
    else:
        x = x.abs()

    if region is None:
        region = [min(targets[0]), min(targets[1]), max(targets[0]), max(targets[1])]

    TGM = th.zeros(x.shape, dtype=th.int8)
    TGM[targets] = 1
    TGM = TGM[region[0]:region[2], region[1]:region[3]].to(x.device)
    x = x[region[0]:region[2], region[1]:region[3]]

    # mask of BG
    BGM = 1 - TGM

    # pixel number of bgs
    NB = BGM.sum()

    R = th.max(x * TGM) / (((1 / NB) * th.sum(x * BGM)) + EPS)

    TBR = 20 * th.log10(R).item()

    if isshow:
        plt.figure
        plt.subplot(131)
        plt.imshow(x)
        plt.subplot(132)
        plt.imshow(TGM)
        plt.subplot(133)
        plt.imshow(x * TGM)
        plt.show()

    return TBR


def tbr2(X, tgrs, subrs=None, isshow=False):
    r"""Target-to-Background Ratio (TBR)

    .. math::
        \begin{align}
        {\rm TBR} = 20{\rm log}_{10}\left(\frac{{\rm max}_{i\in{\mathbb T}}(|{\bf X}_i|)}{{(1/N_{\mathbb B})}\Sigma_{j\in \mathbb B}|{\bf X}_j|)}\right)
        \label{equ:TBR}
        \end{align}

    Parameters
    ----------
    X : tensor
        The input image, if it's complex-valued, it's amplitude is used.
    tgrs : list, optional
        target regions:[[TG1], [TG2], ..., [TGn]], [TGk] = [lefttop, rightbottom]]
    subrs : list, optional
        sub regions:[[SUB1], [SUB2], ..., [SUBn]], [SUBk] = [lefttop, rightbottom]]
    isshow : bool, optional
        show target mask given by :attr:`tgrs` (default: False)

    Returns
    -------
    TBR : float
        Target-to-Background Ratio (TBR)

    Raises
    ------
    TypeError
        tgrs mast be given
    """

    if th.is_complex(X):
        Xabs = X.abs()
    elif X.shape[-1] == 2:
        Xabs = X.pow(2).sum(-1).sqrt()
    else:
        Xabs = X.abs()

    if tgrs is None:
        raise TypeError("Please give the regions of targets!")

    if subrs is not None:
        # mask of Subregion
        SUBM = th.zeros(Xabs.shape, device=Xabs.device, dtype=th.int8)

        for subr in subrs:
            SUBM[subr[0]:subr[2], subr[1]:subr[3]] = 1
        Xabs = Xabs * SUBM

    # mask of TG
    TGM = th.zeros(Xabs.shape, device=Xabs.device, dtype=th.int8)

    for tgr in tgrs:
        TGM[tgr[0]:tgr[2], tgr[1]:tgr[3]] = 1

    # mask of BG
    BGM = 1 - TGM

    # pixel number of bgs
    NB = BGM.sum()

    R = th.max(Xabs * TGM) / (((1 / NB) * th.sum(Xabs * BGM)) + EPS)

    TBR = 20 * th.log10(R).item()

    if isshow:
        plt.figure
        plt.subplot(131)
        plt.imshow(Xabs)
        plt.subplot(132)
        plt.imshow(TGM)
        plt.subplot(133)
        plt.imshow(Xabs * TGM)
        plt.show()

    return TBR


if __name__ == '__main__':

    X = th.zeros((6, 6))
    X = th.rand((6, 6))
    # X = th.rand((6, 6)) + 1j * th.rand((6, 6))

    targets = [[2, 2, 3, 3], [2, 3, 2, 3]]

    X[2:4, 2:4] = 10

    print(X)

    TBR = tbr(X, targets=targets, region=[0, 0, 6, 6])

    print("TBR", TBR)



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : scatter_selection.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import torch as th
import numpy as np
import matplotlib.pyplot as plt


def center_dominant_scatters(x, axis=-2, isplot=False):
    """Center the dominant scatters.

    Parameters
    ----------
    x : tensor
        Complex SAR image tensor.
    axis : int, optional
        The axis to be centered in.
    isplot : bool, optional
        Whether to plot.

    Returns
    -------
    tensor
        Centered.

    Raises
    ------
    TypeError
        Intput must be complex image!
    """
    if th.is_complex(x):
        cplxflag = True
        pass
    elif x.size(-1) == 2:
        cplxflag = False
        x = x[..., 0] + 1j * x[..., 1]
    else:
        raise TypeError('Intput must be complex image!')

    if x.dim() == 2:
        x = x.unsqueeze(0)

    raise(ImportError('Not opened yet!'))


def window_data(z, win=None, axis=-2, isplot=False):
    showidx = -1

    if th.is_complex(z):
        cplxflag = True
        pass
    elif z.size(-1) == 2:
        cplxflag = False
        z = z[..., 0] + 1j * z[..., 1]
    else:
        raise TypeError('Intput must be complex image!')

    if z.dim() == 2:
        z = z.unsqueeze(0)

    N, Na, Nr = z.size()
    Nx = z.size(axis)
    midpos = int(Nx / 2.)
    if axis == -2 or axis == 1:
        axisw, axiso = -2, -1
        winshape = [Nx, 1]
    if axis == -1 or axis == 2:
        axisw, axiso = -1, -2
        winshape = [1, Nx]

    # ---Step 3: Windowing
    ncoh_avg_win = th.sum(z.conj() * z, axis=axiso)  # N-Nx
    ncoh_avg_win_db20 = 20. * th.log10(th.abs(ncoh_avg_win))  # N-Nx
    win_cutoff = th.mean(ncoh_avg_win_db20, axis=-1)  # N-1

    raise(ImportError('Not opened yet!'))



def spotlight_width(H, BWa, Lsar):
    ...

def phase_correction(x, phi, nsub=None, axis=-2, ftshift=True):
    ...

def pgaf_sm_1iter(x, windb=None, est='ML', deg=4, axis=-2, isrmlpe=False, iscrct=False, isplot=False):
    r"""perform phase gradient autofocus 1 iter

    Perform phase gradient autofocus 1 iter as described in [1].

    revised for stripmap SAR

    - [1] C.V. Jakowatz, D.E. Wahl, P.H. Eichel, D.C. Ghiglia, P.A. Thompson,
    {\em Spotlight-mode Synthetic Aperture Radar: A Signal Processing
    Approach.} Springer, 1996.

    - [2] D.E. Wahl, P.H. Eichel, D.C. Ghiglia, C.V. Jakowatz, "Phase
    gradient autofocus-a robust tool for high resolution SAR phase
    correction," IEEE Trans. Aero. & Elec. Sys., vol. 30, no. 3, 1994.

    Args:
        x (Tensor): Complex SAR image :math:`{\bm X}\in {\mathbb C}^{N×N_a×N_r}`, where :math:`N` is the batchsize.
        windb (None or float, optional): Cutoff for window (the default is None, which use the mean as the cutoff.)
        est (str, optional): Estimator, ``'ML'`` for Maximum Likelihood estimation, ``'LUMV'`` for Linear Unbiased Minimum Variance estimation (the default is 'ML')
        deg (int): The degree of polynominal
        axis (int, optional): Autofocus axis.
        isrmlpe (bool, optional): Is remove linear phase error? (the default is False)
        iscrct (bool, optional): Is corrected image? (the default is False)
        isplot (bool, optional): Is plot estimated phase error? (the default is False)

    Returns:
        xc (Tensor): Corrected SAR image :math:`{\bm Y}\in {\mathbb C}^{N×N_a×N_r}`, only if :attr:`iscrct` is ``True``, xc is returned.
        phi (Tensor): Estimated phase error :math:`\phi\in {\mathbb R}^{N×N_a}`.

    Raises:
        TypeError: :attr:`x` is complex and should be in complex or real represent formation!
        ValueError: Not supported estimator! Supported are ``'ML'`` and ``'LUMV'``.

    """

def pgaf_sm(SI, nsar, nsub=None, windb=None, est='ML', deg=4, niter=None, tol=1.e-6, isplot=False, islog=False):
    r"""Phase gradient autofocus for stripmap SAR.

    Phase gradient autofocus for stripmap SAR.

    Args:
        SI (Tensor): Complex SAR image :math:`N_a×N_r`.
        nsar (int): Number of synthetic aperture pixels.
        nsub (int, optional): Number of sub-aperture pixels. (the default is :math:`{\rm min}{N_{sar}, N_a}`)
        windb (None or float, optional): cutoff for window (the default is None, which use the mean as the cutoff.)
        est (str, optional): estimator, ``'ML'`` for Maximum Likelihood estimation, ``'LUMV'`` for Linear Unbiased Minimum Variance estimation (the default is 'ML')
        deg (int): Polynomial degrees (default 4) to fit the error, once fitted, the term of deg=[0,1] will be removed.
            If :attr:`deg` is None or lower than 2, then we do not fit the error with polynominal and not remove the linear trend.
        niter (int, optional): Maximum iters (the default is None, which means using :attr:`tol` for stopping)
        tol (float, optional): Phase error tolerance. (the default is 1.e-6)
        isplot (bool, optional): Plot estimated phase error or corrected image. (the default is False)
        islog (bool, optional): Print log information? (the default is False)

    Returns:
        SI (Tensor): Corrected SAR image :math:`{\bm X}\in {\mathbb C}^{N_a×N_r}`.
        phi (Tensor): Estimated phase error :math:`{\phi}\in {\mathbb R}^{N_a×1}`.

    Raises:
        TypeError: The input is complex and should be in complex or real represent formation!
        ValueError: For stripmap SAR, processing sub aperture should be smaller than synthetic aperture!
    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : sardata.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import h5py


class SarData(object):
    r"""SAR data class.
        rawdata
        image
        store
        read

    """

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise ValueError('name must be a string!')
        self._name = value

    @property
    def rawdata(self):
        return self._rawdata

    @rawdata.setter
    def rawdata(self, value):
        self._rawdata = value

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, value):
        self._image = value

    def store(self, sarplat, file):
        print(file)
        f = h5py.File(file, "w")
        grp = f.create_group("sardata")  # sardata
        grp.create_dataset("name", data=self._name)
        grp.create_dataset("image", data=self._image)
        grp.create_dataset("rawdata", data=self._rawdata)

        grp = f.create_group("sarplat")  # sarplat
        grp.create_dataset("name", data=sarplat.name)

        sensor = sarplat.sensor
        sensor_grp = grp.create_group("sensor")
        for key, value in sensor.items():
            sensor_grp.create_dataset(key, data=value)

        acquis = sarplat.acquisition
        acquis_grp = grp.create_group("acquisition")
        for key, value in acquis.items():
            acquis_grp.create_dataset(key, data=value)

        params = sarplat.params
        params_grp = grp.create_group("params")
        for key, value in params.items():
            params_grp.create_dataset(key, data=value)

        f.close()
        print("sar data has been stored in: ", file)

    def read(self, file):
        r"""Read SAR data file

        Read SAR data file (``.pkl`` or ``.mat``)

        Parameters
        ----------
        file : str
            SAR data file path}

        Returns
        -------
        sardata : Instance SarData
            The SAR raw data.

        sarplat : Instance of SarPlat
            The SAR platform for obtain the SAR data.
        """
        f = h5py.File(file, "r")

        ds_sardata = f["sardata"]
        ds_sarpalt = f["sarplat"]

        self._name = ds_sardata['name'].value
        self._image = ds_sardata['image'].value
        self._rawdata = ds_sardata['rawdata'].value

        ds_sarpalt_sensor = ds_sarpalt["sensor"]
        ds_sarpalt_params = ds_sarpalt["params"]
        ds_sarpalt_acquis = ds_sarpalt["acquisition"]
        sensor = dict()
        for key in ds_sarpalt_sensor.keys():
            sensor[key] = ds_sarpalt_sensor[key].value

        params = dict()
        for key in ds_sarpalt_params.keys():
            params[key] = ds_sarpalt_params[key].value

        acquis = dict()
        for key in ds_sarpalt_acquis.keys():
            acquis[key] = ds_sarpalt_acquis[key].value

        sarplat = ds_sarpalt.copy()

        sarplat.sensor = sensor
        sarplat.params = params
        sarplat.acquisition = acquis

        f.close()

        print("read sar data from: ", file)
        return sarplat
